local testFiltering = false

-- The following code is adapted from r61 of Misspelled, downloaded from
-- http://wow.curse.com/downloads/wow-addons/details/misspelled/download/526785.aspx

if string.utf8len == nil then
  require("libs\\UTF8\\utf8data")
  require("libs\\UTF8\\utf8")
end

--Local reference to Global functions for speed
local string_find = string.find
local table_insert = table.insert
local string_len = string.len
local string_lower =  string.lower
local string_sub = string.sub
local math_min = math.min
local fuzzy = LFMonitor.words.fuzzy
local exact = LFMonitor.words.exact

--split a string, at delimiter into a table
-- Compatibility: Lua-5.1
local split_t
local split_fpat
local split_last_end
local split_s, split_e, split_cap
local utf8 = false

function wordSimilarity(source, target)

	if utf8 == true then
	  string_len = string_utf8len
      string_lower = string_utf8lower
	  string_sub = string_utf8sub
	end

	local m = string_len(source)
	local n = string_len(target)
	local maxDist = math_min(m/2+1,n/2+1,4)

	if math.abs(m-n) >= maxDist then
		return 0
	end

	local rows = {{0}}
	m=m+1
	n=n+1
	for x=1, m do
		rows[x] = {0}
		for y=1,n do
			rows[x][y] = 0
		end
	end

	--tprint(rows)

	for i=1,m do
		rows[i][1] = i
	end

	for j=1,n do
		rows[1][j] = j
	end

	--tprint(rows)
	local subsitutionCost

	for i=2, m do
		for j=2,n do
			if string_sub(source, i, i) == string_sub(target, j, j) then
				subsitutionCost = 0
			else
				subsitutionCost = 1
			end

			rows[i][j] = math_min(	rows[i-1][j] + 1, 					--Insertion
									rows[i][j-1] + 1, 					--Deletion
									rows[i-1][j-1] + subsitutionCost)	--Substitution
		end
	end

	local dist = rows[m][n]

	if dist > maxDist then
		return 0
	else
		return dist/maxDist
	end
end

-- end of adapted code

-- *****************
-- * Filter Engine *
-- *****************

--replace links with the link text
local function unlink(msg)

	local pre, linktext, post = string.match(msg, "([^|]*)|c%x+|H[^|]+|h%[([^|]+)%]|h|r(.*)")
	local unlinked

	pre = pre or ""
	post = post or ""

	if (linktext) then

		unlinked = string.lower(pre .. " " .. linktext .. " " .. unlink(post))

	end

	return unlinked or string.lower(msg)

end

--function to add spaces around {star},{square}, etc so they are parsed correctly
local function formatspecial(msg)

	local out = string.gsub(msg, "{", " {")

	out = string.gsub(out, "}", "} ")

	return out

end

local function doSearch(profTxt, msgTxt, typeTxt, sender, guid, omsg)

	local threshold = 0.75

	local lmsg = unlink(msgTxt)
	local pattern, message, pattern2 = nil, nil, nil
	local matched = false
	local profmatch = 0
	local typematch = 0

	if (cfflag == "CUSTOM") then
		for _, ttxt in ipairs(typeTxt) do
			if (string.match(lmsg, ttxt)) then
				return true
			end
		end
		return false
	end

	pattern = ".*(www%.).*"
	message = string.match(lmsg, pattern)
	matchedvalue = nil

	if (message ~= nil) then

		--found probable spam, so ignore
		matchquality = 0

	end

	if (profTxt[1] ~= nil) then --nop
		profmatch = 1
	else
		profmatch = fuzzyMessageMatch(lmsg, profTxt)
	end
	typematch = fuzzyMessageMatch(lmsg, typeTxt)
	return (profmatch >= threshold) and (typematch >= threshold)
end

-- ***********************
-- * Debug Functions *
-- ***********************

local function debug(msg)

    DEFAULT_CHAT_FRAME:AddMessage(msg, 1, 0, 0)

end

-- **********************************************
-- * Fuzzy matching & classifier implementation *
-- **********************************************

local function StringHash(text)
  local counter = 1
  local len = string_len(text)
  local k
  for i = 1, len, 4 do
    k = string.byte(text,i)*16777216 + (string.byte(text,i+2) or len)*65536 + (string.byte(text,i+2) or len)*256 + (string.byte(text,i+3) or len)
	counter = bit.bxor(math.fmod(bit.band(k, 0x5A5A5A5A), bit.band(k, 0xA5A5A5A5) + 1), k) + math.fmod(counter,4294967291)
  end
  return math.fmod(counter, 4294967279)
end

local function cacheLookup(table, item, fallback, disableCaching)
  if disableCaching then
    return fallback()
  end
--  local n
--  if not table.item then
--    table.item = {}
--	for n=1,table.max do
--	  table.item[n] = false
--	end
--  end
  if not table.score then
    table.score = {}
--	for n=1,table.max do
--	  table.score[n] = false
--	end
  end
--  local hashKey = math.fmod(StringHash(item), table.max) + 1
--  if table.item[hashKey] == item and table.score[hashKey] then
--	debug("cacheLookup found entry "..tostring(table.score[hashKey]).." for item "..item)
--	return table.score[hashKey]
  if table.score[item] then
	return table.score[item]
  else
	local score = fallback()
--	debug("cacheLookup stored result "..tostring(score).." for item "..item)
--	table.item[hashKey] = item
	table.score[item] = score
	if table.count >= table.max then
		table.score[next(table.score)] = nil
	else
		table.count = table.count + 1
	end
	return score
  end
end

local function fuzzyWordMatch(currentWord)
  local fuzzy = fuzzy
  local exact = exact
  local length = string_len(currentWord)
  if length < 3 or length > 15 then
	-- don't match 1- or 2-letter abbrs inexactly, and don't
	  -- waste CPU cycles on words that are too long to match anything
	return {{},{}}
  end
  local matches = {}
  local strengths = {}
  local matchStrength = 0
  for candidateWord, _ in pairs(fuzzy) do
	matchStrength = wordSimilarity(currentWord,candidateWord)
	if matchStrength > 0 then
		tinsert(matches,candidateWord)
		tinsert(strengths,matchStrength)
	end
  end
  return {matches,strengths}
end

local function cachingWordMatch(currentWord, disableCaching)
  if exact[currentWord] then
--    debug("Found exact word match in LFMonitor.words.exact for "..currentWord)
	return {{currentWord},{2}}
  elseif fuzzy[currentWord] then
--    debug("Found exact word match in LFMonitor.words.fuzzy for "..currentWord)
    return {{currentWord},{1}}
  end
  return cacheLookup(LFMonitorWordCache, currentWord, function() return fuzzyWordMatch(currentWord) end, disableCaching)
end

local function classifyMessage(message, disableCaching)

  message = string_lower(formatspecial(unlink(message)))
  local words = {}
  local result = {
	["type"] = {
		["LFPROF"] = 0,
		["LFW"] = 0,
		["LFM"] = 0,
		["LFG"] = 0,
		["WTB"] = 0,
		["WTS"] = 0
	},
	["class"] = {},
	["raid"] = {},
	["role"] = {},
	["prof"] = {}
  }
  local matches = {}
  if testFiltering then
	debug("Message: "..message)
  end
  local wordCount = 0
  local i, match, strength, entry, matchedwords, scores, classtype, classes, class, score, highestScore
  for word in string.gmatch(message, LFMonitor.WORD_PATTERN) do
    wordCount = wordCount + 1
	matchedwords, scores = unpack(cachingWordMatch(word, disableCaching))
	for i=1,#matchedwords do
		matches[matchedwords[i]] = math.max(matches[matchedwords[i]] or 0, scores[i])
	end
  end
  if wordCount >= 2 then
	for match, strength in pairs(matches) do
--	    debug("Looking up "..match.." which matched with strength "..strength)
		if strength >= 2 then
			entry = exact[match]
			strength = 1
		else
			entry = fuzzy[match]
		end
		for classtype, classes in pairs(entry) do
			for class, score in pairs(classes) do
				result[classtype][class] = (result[classtype][class] or 0) + score * strength
			end
		end
	end
	for classtype, classes in pairs(result) do
		highestScore = 1
		for class, score in pairs(classes) do
		  if score > highestScore then
			highestScore = score
		  end
		end
		for class, score in pairs(classes) do
		  result[classtype][class] = 1 - score/highestScore
		  if testFiltering then
			debug(classtype.."="..class..": "..result[classtype][class])
		  end
		end
	end
  end
  return result
end

local function cachingClassifyMessage(message, disableWordCaching)
  return cacheLookup(LFMonitorCache, message, function() return classifyMessage(message, disableWordCaching) end)
end

local function isInDesirableClass(result, desired, sensitivity)
  for class, matches in pairs(result) do
    if matches <= sensitivity and desired[class] == 1 then
	  return true
	end
  end
  return false
end

local function isDesiredProfession(result, sensitivity)
  if LFMonitorDB.monitoring.lf == 1 then
    return true
  end
  local crafts
  if LFMonitorDB.monitoring.misc == 1 then
    crafts = {
		jc = 1,
		enc = 1,
		lw = 1,
		ins = 1,
		tlr = 1,
		bs = 1,
		alc = 1,
		eng = 1,
		misc = 1
	}
  else
    crafts = LFMonitorDB.professions
  end
  return isInDesirableClass(result["prof"], crafts, sensitivity)
end

local function isDesiredRaid(result, sensitivity)
  return LFMonitorDB.options.overrideraid == 1 or isInDesirableClass(result["raid"], LFMonitorDB.raids, sensitivity)
end

local function isDesiredCharacterClass(message, sensitivity)
  return isInDesirableClass(result["class"],LFMonitorDB.classes, sensitivity)
end

local function isDesiredRole(message, sensitivity)
  return isInDesirableClass(result["role"], LFMonitorDB.roles, sensitivity)
end

-- ************************
-- * Variable Declaration *
-- ************************

local alertframe, cfflag, ddmenu, clmenu, pname, pflag, itemlink = nil, nil, nil, nil, nil, nil
local mouse, buttons = false, false
local inactive, combat = false, false
local tradechannel = -1
local _G = _G
local eventframe = nil
local events = {}
local matchedval = nil

-- ***************************
-- * Miscellaneous functions *
-- ***************************
local function sendWhisper(priority, prefix, msg, type, lang, dest)

	local ch, msgchunk, segment, m
	local space = 0
	local msgs = {}

	msgchunk = msg

	if (string.len(msg) > 255) then

		while (string.len(msgchunk) > 255) do

			segment = string.sub(msgchunk, 1, 255)

			for ch = 255, 1, -1 do

				if (string.sub(segment, ch, ch) == " ") then

					space = ch
					break

				end

			end

			if (space == 0) then

				space = 255

			end

			table.insert(msgs, string.sub(segment, 1, space - 1))
			msgchunk = string.sub(msgchunk, space + 1)

		end

	end

	table.insert(msgs, msgchunk)

	for _, m in ipairs(msgs) do

		ChatThrottleLib:SendChatMessage(priority, prefix, m, type, lang, dest)

	end

end

local function resetEvent()

	if (LFMMAIN:GetScript("OnUpdate") == nil) then

		LFMMAIN:SetScript("OnUpdate", function(self, elapsed) LFMonitor:AlertOnUpdate(self, elapsed) end)

	end

end

function LFMonitor:PlayAlert(id)

	local sound = LFMonitor.sounds[id]

	if (LFMonitorDB.options.soundon == 1) then

		if (sound.type == "game") then

			PlaySound(sound.file)

		else

			PlaySoundFile("Interface\\AddOns\\LFMonitor\\Sounds\\" .. sound.file)

		end

	end

end

local function GetAutoReply()

	local reply

	if (matchedvalue) then

		reply = LFMonitorDB.responses[matchedvalue]

	end

	return reply

end

function LFMonitor:ChatCommand(input)

	local command = string.lower(input or "")

	if (command == "show") then

		LFMonitorDB.options.disabled = 0
		SetDisable(false)

	elseif (command == "hide") then

		LFMonitorDB.options.disabled = 1
		SetDisable(true)

	else

		InterfaceOptionsFrame_OpenToCategory(_G.LFMonitorOptionsMain)

	end

end

local function GetLocalTime()

	local format = LFMonitor.timestamps[LFMonitorDB.options.timestamp].format
	local localtime = date(format)

	return localtime

end

local function IsInstance()

	local inInstance, instanceType = IsInInstance()
	local ininstance = false

	if (inInstance ~= nil) then

		if (instanceType ~= "none") then

  			ininstance = true

		end

	end

  	return ininstance

end

local function alertHide()

	alertframe:Hide()

end

local function respond(self, event, ...)

	if (events[event]) then

		events[event](event, ...)

	end

end

function LFMonitor:lfmframeOnUpdate(self, elapsed)

	local flash = LFMEndFlash
	local flashtimer = LFMonitor.frame.flashtimer + elapsed

	if (alertframe:AtBottom()) then

		if (flash:IsShown()) then

			flash:Hide()

		end

		LFMonitor.frame:SetScript("OnUpdate", nil)

	elseif (flashtimer < CHAT_BUTTON_FLASH_TIME ) then

		LFMonitor.frame.flashtimer = flashtimer

	else

		LFMonitor.frame.flashtimer = 0

		if (flash:IsShown()) then

			flash:Hide()

		else

			flash:Show()

		end

	end

end

function LFMonitor:GetChannelNumber(channel)

	local channels = {GetChannelList()}
	local channelnumber = -1

	for num = 1, #channels, 2 do

		if (channels[num + 1] == channel) then

			channelnumber = num

		end

	end

	return channelnumber

end

local function eventframeInit(init)

	if (LFMonitor.testmode == true) then

		eventframe:RegisterEvent("CHAT_MSG_WHISPER_INFORM")

	else

		eventframe:RegisterEvent("CHAT_MSG_CHANNEL")
		eventframe:RegisterEvent("CHAT_MSG_SAY")
		eventframe:RegisterEvent("CHAT_MSG_YELL")

	end

	if (not inactive) then

		eventframe:RegisterEvent("PLAYER_UPDATE_RESTING")
		eventframe:RegisterEvent("PLAYER_ENTERING_WORLD")
		eventframe:RegisterEvent("PLAYER_REGEN_DISABLED")
		eventframe:RegisterEvent("PLAYER_REGEN_ENABLED")
		eventframe:RegisterEvent("ZONE_CHANGED_NEW_AREA")
		eventframe:RegisterEvent("ZONE_CHANGED")
		eventframe:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE")

	end

	inactive = false
	eventframe:SetScript("OnEvent", respond)

	if (init) then

		tradechannel = LFMonitor:GetChannelNumber("Trade")
		LFMonitor:DisableCheck()

	end

	alertframe:Clear()

end

local function SetDisable(disable)

	if (disable == true) then

		if (alertframe:IsVisible()) then

			LFMonitor:OnDisable()
			inactive = true

		end

	else

		if (not alertframe:IsVisible()) then

			UIFrameFadeIn(alertframe, 1)
			resetEvent()
			eventframeInit()
			inactive = false

		end

	end

end

function LFMonitor:DisableCheck()

	local disable = false

	tradechannel = LFMonitor:GetChannelNumber("Trade")

	if (LFMonitorDB.options.hinst == 1) then

		if (IsInstance() == 1) then

			disable = true

		end

	end

	if (LFMonitorDB.options.rested == 1) then

		if (not IsResting()) then

			disable = true

		end

	end

	if (LFMonitorDB.options.trade == 1) then

		if (tradechannel == -1) then

			disable = true

		end

	end

	SetDisable(disable)

end

function LFMonitor:IsInBounds()

	local scale = UIParent:GetEffectiveScale()
	local x, y = GetCursorPosition()
	local xmin = LFMonitorDB.alert.position.x
	local xmax = xmin + LFMonitorDB.alert.position.width
	local ymax = LFMonitorDB.alert.position.y
	local ymin = ymax - LFMonitorDB.alert.position.height
	local inbounds = nil

	x = x / scale
	y = y / scale

	if ((x >= xmin) and (x <= xmax) and (y >= ymin) and (y <= ymax)) then

		inbounds = true

	end

	return inbounds

end

-- *************************
-- * Alert Frame Functions *
-- *************************

function LFMonitor:AlertShowButtons()

	UIFrameFadeIn(ScrollAlertFrameUpBtn, 0.5, 0, 1)
	UIFrameFadeIn(ScrollAlertFrameDownBtn, 0.5, 0, 1)
	UIFrameFadeIn(ScrollAlertFrameEndBtn, 0.5, 0, 1)

end

function LFMonitor:AlertHideButtons()

	ScrollAlertFrameUpBtn:Hide()
	ScrollAlertFrameDownBtn:Hide()
	ScrollAlertFrameEndBtn:Hide()

end

local function checkHistory(msg, sender)

--ignore duplicate messages sent within 20 seconds of each other
--to try and cut down on channel spam

	local i, v
	local valid = 1

	if (#LFMonitor.history > 0) then

		for i,v in pairs(LFMonitor.history) do

			if ((v.msender == sender) and (v.mmsg == msg)) then

				if (GetTime() - v.mtime < 20) then

					valid = nil

				end

			end

		end

	end

	if (valid) then

		if (LFMonitor.hindex == 10) then

			LFMonitor.hindex = 1

		end

		table.insert(LFMonitor.history, LFMonitor.hindex, {msender = sender, mmsg = msg, mtime = GetTime()})
		LFMonitor.hindex = LFMonitor.hindex + 1

	end

	return valid

end

local function addAlert(msg, sender, guid)
	local hlink
	local plink = "[" .. sender .. "]:|h|r "
	local tlink = "|c" .. LFMonitorDB.alert.font.timestamp .. GetLocalTime() .. "|r "
	local mcol = "|c" .. LFMonitorDB.alert.font.text
	local pcol

	if guid == -99999 then

		return

	end

	if (LFMonitorDB.options.useclasscolours == 1) then

		pcol = LFMonitor:GetClassColour(guid)

	else

		pcol = LFMonitorDB.alert.font.player

	end

	hlink = "|c" .. pcol .. "|Hplayer:" .. sender .. ":LFMONITOR:" .. cfflag .. "|h"

	if (checkHistory(msg, sender)) then

		LFMonitor:PlayAlert(LFMonitorDB.options.sound)

		if (alertframe:GetAlpha() ~= 1) then

			LFMonitor:AlertFadeIn(0.1)

		end

		alertframe.TimeSinceLastUpdate = 0
		alertframe:AddMessage(tlink .. hlink .. plink .. mcol .. LFMonitor:FormatMessage(msg) .. "|r", 1, 1, 1, 1, 60.0)

	end

end

function LFMonitor:SetLocked(locked)

	alertframe.isLocked = locked
	LFMonitorDB.options.locked = locked

end

function LFMonitor:AlertFadeIn(ftime)

	if (not mouse) then

		if (alertframe:GetAlpha() ~= 1) then

			UIFrameFadeIn(alertframe, ftime, 0.3, 1)
			resetEvent()

		end

	end

end

function LFMonitor:AlertFadeOut(ftime)

	if (alertframe:GetAlpha() == 1) then

		UIFrameFadeOut(alertframe, ftime, 1, 0.3)

	end

end

function LFMonitor:SetMouse(bValue)

	mouse = bValue

end

function LFMonitor:SetAlertFont()

	local r, g, b, a = LFMonitor:HexToRGB(LFMonitorDB.alert.font.text)
	local fontfile = LFMonitor.fonts[LFMonitorDB.alert.font.index].file

	alertframe:SetFont(fontfile, LFMonitorDB.alert.font.height)

end

function LFMonitor:SetAlertBackdropColour()

	local r, g, b, a = LFMonitor:HexToRGB(LFMonitorDB.alert.backdrop.background)

	alertframe:SetBackdropColor(r, g, b, a)

end

function LFMonitor:SetAlertBorderColour()

	local r, g, b, a = LFMonitor:HexToRGB(LFMonitorDB.alert.backdrop.border)

	alertframe:SetBackdropBorderColor(r, g, b, a)

end

function LFMonitor:ClearAlerts()
	alertframe:Clear()
end

-- *******************
-- * Event Functions *
-- *******************

function LFMonitor:OnInitialise()

	local x, y
	local slashcmd, slashidx = "LFMONITOR_CMD", 1

	SLASH_LFMONITOR_CMD1 = "/" .. LFMonitor.CHAT_COMMANDS[1]
	SLASH_LFMONITOR_CMD2 = "/" .. LFMonitor.CHAT_COMMANDS[2]
	SlashCmdList["LFMONITOR_CMD"] = function(input) LFMonitor:ChatCommand(input) end

	LFMDBSVARS = LFMonitorDB

	if (IsAddOnLoaded("TradeForwarder")) then

		LFMonitor.tfw = true

	else

		LFMonitor.tfw = nil

	end

	--remove old saved data that is no longer required
	if (LFMonitorDB.craftlist) then

		LFMonitorDB.craftlist = nil

	end

end

function LFMonitor:OnDisable()

	local r, g, b, a = LFMMAIN:GetBackdropColor()

	if (LFMonitor.testmode == true) then

		eventframe:UnregisterEvent("CHAT_MSG_WHISPER_INFORM")

	else

		eventframe:UnregisterEvent("CHAT_MSG_CHANNEL")
		eventframe:UnregisterEvent("CHAT_MSG_SAY")
		eventframe:UnregisterEvent("CHAT_MSG_YELL")

	end

	LFMonitor.frame:SetScript("OnUpdate", nil)
	LFMMAIN:SetScript("OnUpdate", nil)

	UIFrameFadeOut(alertframe, 1, a, 0)
	alertframe.fadeInfo.finishedFunc = alertHide
	inactive = true

end

function events.PLAYER_REGEN_DISABLED()

	--player entering combat
	if (LFMonitorDB.options.combat == 1) then

		if (alertframe:IsVisible()) then

			dis = true
			combat = true
			UIFrameFadeOut(alertframe, 1)
			alertframe.fadeInfo.finishedFunc = alertHide

		end

	end

end

function events.PLAYER_REGEN_ENABLED()

	--player leaving combat
	if (combat) then

		UIFrameFadeIn(alertframe, 1)
		resetEvent()
		combat = false

	end

end

function events.PLAYER_UPDATE_RESTING(event, ...)

	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

end

function events.ZONE_CHANGED_NEW_AREA(event, ...)

	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

end

function events.ZONE_CHANGED(event, ...)

	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

end

function events.PLAYER_ENTERING_WORLD(event, ...)

	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

	if (LFMonitorDB.raids.eoe) then

		LFMonitorDB.raids = {}

	end

end

function LFMonitor:AlertOnUpdate(self, elapsed)

	self.TimeSinceLastUpdate = self.TimeSinceLastUpdate + elapsed

	if (self.TimeSinceLastUpdate < LFMonitor.LFM_Update_Int) then

		return

	end

	if (alertframe:IsVisible()) then

		if (alertframe:GetAlpha() == 1) then

			if (mouse == false) then

				self.TimeSinceLastUpdate = 0
				LFMonitor:AlertHideButtons()
		    	LFMonitor:AlertFadeOut(2)
		    	LFMonitor:AlertFadeOut(2)
				LFMMAIN:SetScript("OnUpdate", nil)
				mouse = false

		  	end

		end

	end

end

function LFMonitor:OnHyperlinkShow(self, link, text, button)

	SetItemRef(link, text, button)

end

local function tableEmpty(table)

	for _, _ in pairs(table) do
		return false
	end
	return true

end

local function ChatParser(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, _, counter, guid, ...)

	local sensitivity = LFMonitorDB.options.sensitivity
	local message = nil
	local pattern, lmsg = nil, nil
	local omsg, classtxt = nil, nil
	local lfmclasses = {"dk", "dr", "hu", "ma", "pa", "pr", "ro", "sh", "wa", "wr"}
	local profs = {"jc", "enc", "lw", "ins", "tlr", "bs", "alc", "eng", "misc"}
	local misc = {"lfg", "wtb", "lf", "wts", "lfw"}
	local matched, lfmgen = false, true
	local channelmatch = false
	local tmp = {}
	local i, channels = {}
	local tfwidx, tfwname, tfwmsg, tfwcolour
	local lfmsender, original

	lfmsender = sender

	if (LFMonitorDB.channels.general == 1) then

		if ((zoneid == 1) or (event == "CHAT_MSG_SAY") or (event == "CHAT_MSG_YELL")) then

			channelmatch = true

		end

	end

	if (LFMonitorDB.channels.trade == 1) then

		if (zoneid == 2) then

			channelmatch = true

		end

	end

	if (LFMonitorDB.channels.guild == 1) then

		if (event == "CHAT_MSG_GUILD") then

			channelmatch = true

		end

	end

	if (LFMonitorDB.channels.lfg == 1) then

		if (zoneid == 26) then

			channelmatch = true

		end

	end

	if (channelmatch == false) then

		channels = {GetChannelList()}

		for i = 1, #channels, 2 do

			if (channels[i] > 4) then

				if (LFMonitorDB.channels["cc" .. channels[i + 1]] == 1) then

					if (channelnumber == channels[i]) then

						channelmatch = true

					end

				end
			end

		end

	end

	if (LFMMAIN:IsVisible()) then

		if ((channelmatch == true) or (LFMonitor.testmode == true)) then

			original = chatmessage
			omsg = chatmessage

			if LFMonitor.tfw then

				tfwidx, tfwname, tfwmsg, tfwcolor = strmatch(chatmessage, LFMonitor.TFW_PATTERN)

				if tfwmsg then

					omsg = tfwmsg
					lfmsender = tfwname

				end

			end

			local result = cachingClassifyMessage(omsg, messageTypes, "type", false)

			if result["type"]["WTB"] <= sensitivity and LFMonitorDB.monitoring.wtb == 1 then
				cfflag = "WTB"
				matched = true
			elseif result["type"]["WTS"] <= sensitivity and LFMonitorDB.monitoring.wts == 1 then
				cfflag = "WTS"
				matched = true
			elseif result["type"]["LFW"] <= sensitivity and LFMonitorDB.monitoring.lfw == 1 then
				cfflag = "LFW"
				matched = true
			end
			if (not matched) and result["type"]["LFPROF"] <= sensitivity then
				cfflag = "LFPROF"
				matched = isDesiredProfession(result, sensitivity)
			end
			if (not matched) and result["type"]["LFG"] <= sensitivity and LFMonitorDB.monitoring.lfg == 1 then
				cfflag = "LFG"
				matched = isDesiredRaid(result, sensitivity)
			end
			if (not matched) and result["type"]["LFM"] <= sensitivity and LFMonitorDB.monitoring.lfm == 1 then
				cfflag = "LFM"
				matched = isDesiredRaid(result, sensitivity) and
				          (isDesiredCharacterClass(result, sensitivity) or isDesiredRole(result, sensitivity))
			end
			if (not matched) then

				if ((LFMonitorDB.options.basic or "") ~= "") then
					omsg = formatspecial(unlink(omsg))

					cfflag = "CUSTOM"
					tmp = {}

					if (string.find(LFMonitorDB.options.basicesc, " ")) then

						tmp = {strsplit(" ", string.lower(LFMonitorDB.options.basicesc))}

					else

						table.insert(tmp, string.lower(LFMonitorDB.options.basicesc))

					end

					matched = doSearch(LFMonitor.nop, omsg, tmp, lfmsender, guid, original)

				end

			end

			if (not matched) then

				if ((LFMonitorDB.options.adv or "") ~= "") then
					omsg = formatspecial(unlink(omsg))

					cfflag = "CUSTOM"
					tmp = {}
					table.insert(tmp, LFMonitorDB.options.adv)
					matched = doSearch(LFMonitor.nop, omsg, tmp, lfmsender, guid, original)

				end

			end
			if matched then
				addAlert(original, lfmsender, guid)
			end

		end

	end

end

function events.CHAT_MSG_CHANNEL(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

	ChatParser(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

end

function events.CHAT_MSG_WHISPER_INFORM(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

	ChatParser(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

end

function events.CHAT_MSG_YELL(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

	ChatParser(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

end

function events.CHAT_MSG_SAY(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

	ChatParser(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

end

function events.CHAT_MSG_CHANNEL_NOTICE(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

end

function LFMonitor:OnEnable()

	LFMonitor:CreateOptions()

	if (LFMonitorDB.alert.position.x == -1) then

		x = (UIParent:GetWidth() - LFMonitorDB.alert.position.width) / 2
		y = (UIParent:GetHeight() - LFMonitorDB.alert.position.height) / 2
		LFMonitorDB.alert.position.x = x
		LFMonitorDB.alert.position.y = y

	end

	alertframe = LFMonitor:CreateAlertFrame()

	eventframe = CreateFrame("Frame", "LFMonitorEventFrame")
	eventframeInit(true)
	LFMonitor:BuildTables()
	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

	if (IsAddOnLoaded("TradeForwarder")) then

		LFMonitor.tfw = 1

	else

		LFMonitor.tfw = nil

	end

end

-- ******************
-- * Dropdown Menus *
-- ******************

local function LFMonitor_DDMenuItem_OnClick(self, button, down)

	local opt = self.value
	local reply, rtmp = nil, nil
	local cl = nil

	PlaySound("igMainMenuOption")

	if (opt == 1) then

		--autoreply
		reply = GetAutoReply()

		if (reply == nil or reply == "") then

			--nothing to autoreply with
			StaticPopup_Show("LFMONITOR_NOAUTOREPLY")

		else

			rtmp = string.sub(reply, 1, 255)

			sendWhisper("NORMAL", "LFMONITOR", reply, "WHISPER", nil, pname)

		end

	elseif(opt == 2) then

		--whisper
		ChatFrame1EditBox:SetText(LFMonitor.WHISPER_COMMAND .. " " .. pname .. " ")

	elseif(opt == 3) then

		--ignore
		AddIgnore(pname)

	elseif(opt == 4) then

		--cancel
		--do nothing

	end

end

local function LFMonitor_DD_Initialise()

	local info

	info = {}
	info.text = pname
	info.value = 0
	info.owner = ddmenu
	info.isTitle = true
	info.func = nil
	info.notCheckable = 1
	UIDropDownMenu_AddButton(info)

	info = {}
	info.text = LFMonitor.AUTOREPLY
	info.value = 1
	info.owner = ddmenu
	info.notCheckable = 1
	info.isTitle = nil

	if (pflag == "LFG" or pflag == "WTB") then

		info.disabled = true

	else

		info.disabled = nil

	end

	info.tooltipTitle = LFMonitor.AUTOREPLY
	info.tooltipText = string.format(LFMonitor.SEND_AUTO, pname)
	info.func = LFMonitor_DDMenuItem_OnClick
	UIDropDownMenu_AddButton(info, 1)

	info = {}
	info.text = _G.WHISPER
	info.value = 2
	info.owner = ddmenu
	info.isTitle = nil
	info.disabled = nil
	info.notCheckable = 1
	info.tooltipTitle = _G.WHISPER
	info.tooltipText = string.format(LFMonitor.WHISPER, pname)
	info.func = LFMonitor_DDMenuItem_OnClick
	UIDropDownMenu_AddButton(info, 1)

	info = {}
	info.text = _G.IGNORE
	info.value = 3
	info.owner = ddmenu
	info.isTitle = nil
	info.disabled = nil
	info.tooltipTitle = _G.IGNORE
	info.notCheckable = 1
	info.tooltipText = string.format(LFMonitor.IGNORE_LIST, pname)
	info.func = LFMonitor_DDMenuItem_OnClick
	UIDropDownMenu_AddButton(info, 1)

	info = {}
	info.text = _G.CANCEL
	info.value = 4
	info.owner = ddmenu
	info.isTitle = nil
	info.disabled = nil
	info.notCheckable = 1
	info.tooltipTitle = _G.CANCEL
	info.tooltipText = LFMonitor.CLOSE_MENU
	info.func = LFMonitor_DDMenuItem_OnClick
	UIDropDownMenu_AddButton(info, 1)

end

local function LFMonitor_ShowDDMenu(link)

	local htype, hname, hmod, flag = strsplit(":", link)

	pname = hname
	pflag = flag

	ddmenu = _G.LFMonitorDDMenu
	ddmenu.initialize = LFMonitor_DD_Initialise
	ddmenu.displayMode = "MENU"
	ddmenu.name = "LFMonitor"
	PlaySound("igMainMenuOpen")
	ToggleDropDownMenu(1, nil, ddmenu, "cursor")

end

-- *********
-- * Hooks *
-- *********

--Hook SetItemRef so LFMonitor can handle its own custom links
local ORIG_SetItemRef = _G.SetItemRef

local function LFMonitor_SetItemRef (link, text, button, ...)

	local htype, hname, hmod, flag = strsplit(":", link)
	local runorig = nil

	if (IsModifiedClick("CHATLINK")) then

		runorig = true

	elseif (button == "RightButton") then

		if (hmod == "LFMONITOR") then

			LFMonitor_ShowDDMenu(link)

		else

			runorig = true

		end

	else

		runorig = true

	end

	if (runorig) then

		ORIG_SetItemRef(link, text, button, ...)

	end

end

_G.SetItemRef = LFMonitor_SetItemRef

--Hook ChatEdit_GetActiveWindow so LFMonitor can capture links in the options frame
local orig_ChatEdit_GetActiveWindow = ChatEdit_GetActiveWindow

function ChatEdit_GetActiveWindow()

    return LFMonitor.EditFocus or orig_ChatEdit_GetActiveWindow()

end

local function benchmark(operation, message, benchmarkCount)
	collectgarbage()
	debugprofilestart()
	for i=1,benchmarkCount do
		operation()
	end
	local run = debugprofilestop()
	local memory = collectgarbage("count")
	debugprofilestart()
	collectgarbage()
	local gc = debugprofilestop()
	local memory_nogarbage = collectgarbage("count")
	debug(string.format("%s: %d ms (%d run, %d gc), %d-%d kB",message,gc+run,run,gc,memory_nogarbage,memory))
end

function randomWord(dictionary, chanceDictionary, chanceLink, chanceSpecial, avgLength)
	local roll = math.random()
	if roll < chanceLink then
		return "|c" .. randomWord(dictionary,0,0,0,8) .. "|H" .. randomWord(dictionary,0,0,0,12) .. "|h[" .. randomWord(dictionary, chanceDictionary, 0, 0, avgLength) .. "]|h|r"
	elseif roll < chanceLink + chanceSpecial then
		return "{" .. randomWord(dictionary, chanceDictionary, 0, 0, avgLength) .. "}"
	elseif roll < chanceLink + chanceSpecial + chanceDictionary then
		return dictionary[math.random(#dictionary)]
	else
		local word = ""
		local chanceStop = 1/(avgLength - 1)
		repeat
			word = word .. strchar(math.random(96,122))
		until math.random() < chanceStop
		return word
	end
end

function randomDict() -- generates list of all keys in LFMonitor.words.exact and LFMonitor.words.fuzzy plus 1000 random new words, for random message generation
	local dictionary = {}
	for k, _ in pairs(LFMonitor.words.exact) do
		tinsert(dictionary, k)
	end
	for k, _ in pairs(LFMonitor.words.fuzzy) do
		if not LFMonitor.words.exact[k] then
			tinsert(dictionary, k)
		end
	end
	for k=1,1000 do
		tinsert(dictionary, randomWord(nil,0,0,0,6))
	end
	return dictionary
end

function LFMonitor:benchmarkStringHash(avgLength, benchmarkCount)
	local strings = {}
	local i
	for i=1,benchmarkCount do
		strings[i] = randomWord(nil,0,0,0,avgLength)
	end
	i=1
	benchmark( function()
			StringHash(strings[i])
			i=i+1
		end, "StringHash", benchmarkCount)
end

function LFMonitor:benchmarkWordMatch(avgLength, benchmarkCount)
	local strings = {}
	local i
	for i=1,benchmarkCount do
		strings[i] = randomWord(nil,0,0,0,avgLength)
	end
	i=1
	benchmark( function()
			fuzzyWordMatch(strings[i])
			i=i+1
		end, "fuzzyWordMatch", benchmarkCount)
end

function LFMonitor:benchmarkWordSimilarity(avgLength, benchmarkCount)
	local strings = {}
	local i
	for i=1,benchmarkCount do
		strings[i] = randomWord(nil,0,0,0,avgLength)
	end
	i=1
	benchmark( function()
			wordSimilarity(strings[i], strings[math.random(benchmarkCount)])
			i=i+1
		end, "fuzzyWordMatch", benchmarkCount)
end

function LFMonitor:benchmarkFiltering(wordCacheSize, messageCacheSize, benchmarkCount)
	local dict = randomDict()
	local avgWordLength = 6
	local avgMessageLength = 6
	local chanceStop = 1/(avgMessageLength - 1)
	local chanceLink = .1 / avgMessageLength -- 1 link per 10 messages
	local chanceSpecial = .02 / avgMessageLength -- 1 special per 50 messages
	local messages = {}
	local n
	for n=1,benchmarkCount do
		messages[n] = ""
		repeat
			messages[n] = messages[n] .. randomWord(dict, .9, chanceLink, chanceSpecial, avgWordLength) .. " "
		until math.random() < chanceStop
	end
	LFMonitorCache = {score = {}, count=0, max = messageCacheSize}
	LFMonitorWordCache = {score = {}, count=0, max = wordCacheSize}
	local i=1
--	cachingClassifyMessage("foo") -- force cache initialization before the benchmark
--	benchmark( function ()
--		classifyMessage(messages[i], true)
--		i=i+1
--		end, "No caching", benchmarkCount)
--	i=1
--	benchmark( function ()
--		cachingClassifyMessage(messages[i], true)
--		i=i+1
--		end, "New messages, no word caching", benchmarkCount)
	i=1
--	benchmark( function ()
--		classifyMessage(messages[i], false)
--		i=i+1
--		end, "Word caching, no message caching", benchmarkCount)
	LFMonitorWordCache = {score = {}, count = 0, max = wordCacheSize}
--	cachingClassifyMessage("foo") -- force cache initialization before the benchmark
	benchmark( function ()
		cachingClassifyMessage(messages[i], false)
		i=i+1
		end, "New messages, word and message caching", benchmarkCount)
	benchmark( function ()
		cachingClassifyMessage(messages[math.random(benchmarkCount)], true)
		end, "Repeat messages", benchmarkCount)
	LFMonitorCache = {score = {}, count = 0, max = messageCacheSize}
end
